//#define ATAPI_DEBUG
/*
 * MP3 Player, ATAPI support.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * Contact: paul@pjrc.com
 *
 * File		: atapi.c
 * Author	: MBA, the flying dutchman from Gouda, NL.
 * Created	: dec, 2001
 *
 */

// ======================================================================
//
// ========== INCLUDES
//
// ======================================================================
#include "8051.h"
#include "Ata.h"
#include "Atapi.h"
#include "AtapiP.h"
#include <string.h>
#include "paulmon2.h"
#include "as31glue.h"
#include "startup.h"
#include "printf.h"
#include "display.h"
#include "malloc.h"
#include "stricmp.h"
#include "paulmon2.h"

#pragma SAVE
// I do not know exactly what the next pragma NOGCSE does, but it seems to prevent the compiler
// from creating a lot of SLOCs. Some functions needed about 50 SLOCs without this pragma, and 
// only 1 or 2 SLOCS with the pragma.
#pragma NOGCSE

struct REG_CMD_INFO reg_cmd_info;

// ======================================================================
//
// ========== MODULE GLOBAL VARIABLES
//
// ======================================================================

static xdata unsigned char                 m_aucSectorBuffer[2048];
static xdata unsigned char                 m_aucCommandBuffer[2048];
static xdata AtapiDevicePropertiesType     m_tDeviceProperties[ATA_DEVICE_COUNT_MAX];
static xdata _AtapiVolumeDescriptorSet     m_vds[ATA_DEVICE_COUNT_MAX];
static xdata volatile _AtapiDirStructType  m_AtapiDirList[ATA_DEVICE_COUNT_MAX];
static xdata volatile _AtapiFileStructType m_AtapiFileList[ATA_DEVICE_COUNT_MAX];
static xdata volatile simm_id              m_sidAtapiDirListIterator[ATA_DEVICE_COUNT_MAX];
static xdata volatile simm_id              m_sidAtapiFileListIterator[ATA_DEVICE_COUNT_MAX];
static xdata volatile unsigned char        m_ucAtapiDirListSize[ATA_DEVICE_COUNT_MAX];
static xdata volatile unsigned char        m_ucAtapiFileListSize[ATA_DEVICE_COUNT_MAX];
static xdata volatile unsigned char        m_ucAtapiDirListCounter[ATA_DEVICE_COUNT_MAX];
static xdata volatile unsigned char        m_ucAtapiFileListCounter[ATA_DEVICE_COUNT_MAX];


// ======================================================================
//
// ========== PRIVATE FUNCTIONS
//
// ======================================================================

/*********************
 * _AtapiDelay100ms
 *	Private function that issues a delay of 100msec.
 *	The ATAPI specification describes that between certain commands and/or
 *	checks, a 100msec delay must be issued.
 *
 * NOTE:
 *	This delay is hardly used in this implementation. It seems not necessary
 *	in most cases, due to C code latency (I think). However, it will be added
 *	to places where the absence of a required delay causes trouble !!
 */
void _AtapiDelay100ms(void)

{
	_asm
        ;delay for 10*10=100 ms
        ;100ms at 7.3728 MHz is 61440 machine cycles
        ;61440 = 256 * 240
        push	ar0
        push	ar1
	mov     r0, #0
d10ms:	mov     r1, #118	;1 cycle
dd:	djnz    r1,dd		;118 * 2 = 236 cycles
        nop			;1 cycle
	djnz    r0, d10ms	;2 cycles
	pop	ar1
	pop	ar0
	_endasm;
}

/*********************
 * _AtapiValidateName
 *	Private function that validates the ATAPI file name:
 *	- It ignores the part after ";"
 *	- It "converts" the UNICODE string to a normal string by skipping every 0 byte.
 *
 * NOTE:
 *	This routine is likely to be rewritten in (inline) assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 */
// Optimized @31/03/2002 by replacing int's into unsigned chars --> 25% smaller code: 133 instead of 182 bytes...
void _AtapiValidateName(STRING_Z* p_szName, unsigned char p_ucLen)
{
  // ucIterD is declared volatile, to prevent an SDCC bug to happen
  volatile unsigned char ucIterD = 0;
           unsigned char ucIterS = 0;
  
  for (ucIterS = 0; ucIterS < p_ucLen; ucIterS++)
  {
    if (p_szName[ucIterS] == ';')
      break;
      
    if (p_szName[ucIterS])
    {
      p_szName[ucIterD] = p_szName[ucIterS];
      // NOTE: Next increment does NOT work, if ucIterD not declared volatile.
      ucIterD++;
    }
  }
  p_szName[ucIterD] = 0;
}

/*********************
 * _CLong
 *	Private helper function to convert 4 bytes into a long. Byteordering: LSB----MSB (Intel format)
 *
 * NOTE:
 *	This routine is likely to be rewritten in assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 */
unsigned long _CLong (xdata unsigned char *p_ucBytes)
{
  unsigned long ul;

  ul = *((long *)(p_ucBytes));
  return(ul);
}

/*********************
 * _CInt
 *	Private helper function to convert 2 bytes into an int. Byteordering: LSB----MSB (Intel format)
 *
 * NOTE:
 *	This routine is likely to be rewritten in assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 */
unsigned int _CInt (unsigned char *p_ucBytes)
{
  return *((int *)(p_ucBytes));
}

/*********************
 * _Long2Char
 *	Private helper function to convert a long to a 4-byte array. Byteordering: LSB----MSB (Intel format)
 *
 * NOTE:
 *	This routine is likely to be rewritten in assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 */
void _Long2Char (unsigned long p_ulValue, unsigned char *p_acBytes)
{
  *(p_acBytes    ) = (p_ulValue & 0x000000ff);
  *(p_acBytes + 1) = (p_ulValue & 0x0000ff00) >> 8;
  *(p_acBytes + 2) = (p_ulValue & 0x00ff0000) >> 16;
  *(p_acBytes + 3) = (p_ulValue & 0xff000000) >> 24;
}

/*********************
 * _AtapiDebugPrintSectorBuffer
 *	Private helper function that dumps the contents of the SectorBuffer
 *
 */
void _AtapiDebugPrintSectorBuffer()
{
// def out. Need the code space !!!!!!!!!!
// NOTE NOTE NOTE
// If I put the comment start AFTER the var declarations, the compiler issues an error,
// because uiWordCnt is declared twice (indeed, also in another function)!!!!!!!
// AIN'T THAT SOMETHING WIERD ?????
/*
  unsigned int         uiWordCnt = 1024;
  xdata unsigned char* paucBuffer;
  unsigned char        ucLo;
  unsigned char        ucHi;
  unsigned int         uiCounter = 0;
  xdata unsigned int   uiData;
  
/*
  for (uiWordCnt = 0; uiWordCnt < 1024; uiWordCnt++)
  {
#if 1
    ucLo = m_aucSectorBuffer[2 * uiWordCnt];
    ucHi = m_aucSectorBuffer[2 * uiWordCnt + 1];
#else
    ucLo = *(paucBuffer    );
    ucHi = *(paucBuffer + 1);
    paucBuffer = paucBuffer + 2;
#endif
    uiCounter++;
    // print_hex16 seems to flip hi/lo bytes. So swap again...
    uiData = (((unsigned int)(ucLo)) << 8) | (unsigned int)ucHi;
    print_hex16(uiData);
    if ((uiCounter % 10) == 0) print_crlf();
  }
*/
}

/*********************
 * _AtapiGetVdsData
 *	Private function that fetches relevant fields from a Volume Descriptor Structure from CD.
 *	The VD might be one of the following (no more choices available):
 *	- Primary Volume Descriptor, at least 1 (PVD)
 *	- Secundary Volume Descriptor, zero or more (SVD)
 *	- Volume Partition Descriptor (VPD, not used by this implementation)
 *	- Boot (not used by this implementation).
 *
 *	Only those PVD/SVD fields that are relevant for this implementation are copied, meaning
 *	about 30 fields are not used, interpreted or whatever. Lucky us, because Reading all
 *	these fields did cause a lot of trouble (SLOCs, stack etc.)
 *	
 */
void _AtapiGetVdsData(_AtapiVolumeDescriptorStructTypePtr p_pCurVds, unsigned char xdata *p_aucBuffer) reentrant
{
  p_pCurVds->ucVolumeDescriptorType = *(p_aucBuffer);
  p_pCurVds->ucFileSystemFormat     = *(p_aucBuffer +  90);
  p_pCurVds->uiLogicalBlockSize     = _CInt(p_aucBuffer + 128);
  p_pCurVds->ulPathTableSize	    = _CLong(p_aucBuffer + 132);
             printf("PathtableSize = ");
             print_hex32(p_pCurVds->ulPathTableSize);
             print_crlf();
  p_pCurVds->ulLocationLPathTable   = _CLong(p_aucBuffer + 140);
             printf("PathtableLocation = ");
             print_hex32(p_pCurVds->ulLocationLPathTable);
             print_crlf();
  p_pCurVds->dirRootDir.ulDataLen   = _CLong(p_aucBuffer + 166);
  p_pCurVds->ucStructureVersion     = *(p_aucBuffer + 881);
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS
//
// ======================================================================

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - Directory list "class"
//
// ======================================================================

/*********************
 * AtapiDirListIterateInit
 *	Public function that initializes the iterator values before
 *	AtapiDirListIterateNext() can be called.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *
 */
void AtapiDirListIterateInit(unsigned char p_ucDevice)
{
  m_sidAtapiDirListIterator[p_ucDevice] = m_AtapiDirList[p_ucDevice].sidNext;
  m_ucAtapiDirListCounter[p_ucDevice]   = 0;
//  printf("DirInit: sid = %4lx\r\n", m_sidAtapiDirListIterator[p_ucDevice]);
}

/*********************
 * AtapiDirListIterateNext
 *	Public function that fetches the next directory in the DirList and
 *	its parent for given device.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *
 *	Returns true if a directory found, otherwise if list is empty returns false
 *
 */
// Optimized @31/03/2002. Change int's into unsigned char, saving 30 bytes of code...
// Note that this assumes that there are no more than 255 directories on the CD !!!!!
// I'm afraid that this has to change with the 4.7GB DVD's...
unsigned char AtapiDirListIterateNext(unsigned char p_ucDevice,
                                      simm_id*      p_psidDirEntry,
                                      simm_id*      p_psidDirParent)
{
  volatile simm_id     sidDirCur;
  unsigned char        ucPDN;
  simm_id              sidDir;
  _AtapiDirStructType* pDir;
  unsigned char        ucLocator;
  
  sidDirCur = m_sidAtapiDirListIterator[p_ucDevice];
  printf("DirNext: sidDir = %4lx", sidDirCur);
  
  // Check for end of list & return if so.
  if (sidDirCur == 0)
  {
    print_crlf();
    return false;
  }
  
  // Nice and easy.
  // All of a sudden, the last sid gets corrupted somehow by having the value 0xFFFF00
  // instead of 0. Safeguard this "feature" by adding a counter, and listsize...
  // THIS SHOULD NOT BE NECESSARY!
  if ((m_ucAtapiDirListCounter[p_ucDevice]++) == m_ucAtapiDirListSize[p_ucDevice])
  {
    print_crlf();
    return false;
  }
    
  // Return data
  *p_psidDirEntry = sidDirCur;
  
  // Now the difficult one. Locate parent, and return that one too
  // First of all, get the parent directory number for the current directory
  
  ucPDN = (unsigned char)((_AtapiDirStructType*)(addrC(sidDirCur)))->iPDN;
  pDir = &m_AtapiDirList[p_ucDevice];
  for (ucLocator = 0; ucLocator < ucPDN; ucLocator++)
  {
    sidDir = pDir->sidNext;
  }
  *p_psidDirParent = sidDir;
  printf(", sidParent = %4lx\r\n", sidDir);
  
  // Point to next in list
  m_sidAtapiDirListIterator[p_ucDevice] = ((_AtapiDirStructType*)(addrC(sidDirCur)))->sidNext;
  return true;
}

/*********************
 * AtapiDirListUserDataSet
 *	Public function that sets userdata for the current directory.
 *	This data will be fetched later on with the xxxGet function.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiDirListUserDataSet(simm_id       p_sidDirEntry,
                             unsigned long p_ulUserData)
{
  ((_AtapiDirStructType*)addrC(p_sidDirEntry))->ulUserData = p_ulUserData;
//  printf("UserSet: dir = %4lx, data = %4lx\r\n", p_sidDirEntry, p_ulUserData);
}

/*********************
 * AtapiDirListUserDataGet
 *	Public function that gets userdata for the current directory.
 *	This data was previously set with the xxxSet function.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiDirListUserDataGet(simm_id        p_sidDirEntry,
                             unsigned long* p_pulUserData)
{
  *p_pulUserData = ((_AtapiDirStructType*)addrC(p_sidDirEntry))->ulUserData;
//  printf("UserGet: dir = %4lx, data = %4lx\r\n", p_sidDirEntry, *p_pulUserData);
}

/*********************
 * AtapiDirListCreate
 *	Public function that creates the full directory list for the current CD
 *	It traverses the CD to fetch the ATAPI specific directory list and puts them
 *	in a module global list, which is iterated later on with the previous functions.
 *
 *	This list will be used by the DirList/PlayList modules to retrieve the data &
 *	build the directory and playlists. This list therefore is a temporary list which
 *	will be deleted after its use.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiDirListCreate(unsigned char p_ucDevice)

{
  static xdata unsigned char       *paucSectorBuffer;
  static unsigned char             *tempPtr;
  static unsigned char              ucLenDI = 0;
  static xdata _AtapiDirStructType* pptDirRoot;
  static xdata _AtapiDirStructType* pptDirCur;
  static xdata simm_id              sidDirNew;
  static xdata _AtapiDirStructType* pptDirNew;
  static xdata unsigned long        ulPathTableSize = 0;
  static xdata unsigned long        ulPathTableExtent = 0;
  static xdata unsigned char        ucPathTableExtentCount = 0;
  volatile unsigned char            ucDirCount = 0;
  
  printf("DirListCreate++\r\n");
  // If an SVD is defined, use this SVD to build-up the list of directories
  // Normally, only the PVD is used, but in case of Joliet, the SVD is the one we need to get nice
  // long file name support...
  
  if (m_vds[p_ucDevice][_VDSSVD].ucVolumeDescriptorType != 0)
  {
    ulPathTableSize   = m_vds[p_ucDevice][_VDSSVD].ulPathTableSize;
    ulPathTableExtent = m_vds[p_ucDevice][_VDSSVD].ulLocationLPathTable;
  }
  else
  {
    ulPathTableSize   = m_vds[p_ucDevice][_VDSPVD].ulPathTableSize;
    ulPathTableExtent = m_vds[p_ucDevice][_VDSPVD].ulLocationLPathTable;
  }
  
  // Calculate the number of pathtable extents (2048 byte sectors) to read to build the full directory list
  ucPathTableExtentCount = (ulPathTableSize + 2048L) / 2048L;
  printf("PathTableExtentCount = %d\r\n", ucPathTableExtentCount);

  // Setup pointer to DirList to be created by this function
  pptDirRoot = &m_AtapiDirList[p_ucDevice];
  pptDirCur  = pptDirRoot;
  
  printf("Read LBA's PT\r\n");
  while (ucPathTableExtentCount-- > 0)
  {
    AtapiReadSector2(p_ucDevice, ulPathTableExtent); //ucExtent[0], ucExtent[1], ucExtent[2], ucExtent[3]);
    ulPathTableExtent++; //ucExtent[0]++;
  
    // Format of Path Table Record:
    // ----------------------------
    // _BP			Field Name					Content
    // ==			==========					=======
    // 1			Length of Directory Identifier (LEN_DI)		numerical value
    // 2			Extended Attribute Record Length		numerical value
    // 3 to 6			Location of Extent				numerical value
    // 7 to 8			Parent Directory Number				numerical value
    // 9 to (8 + LEN_DI)	Directory Identifier				d-chars, d1-chars, (00) byte
    // (9 + LEN_DI)		Padding Field					(00) byte
  
    paucSectorBuffer = m_aucSectorBuffer;
    // Loop until (not documented) Parent Directory Number = 0 (should start with 1 !!)
 
    printf("Printing dir found...\r\n");
    while (paucSectorBuffer[_BP7] != 0)
    {
      ucDirCount++;
      ucLenDI = paucSectorBuffer[_BP1];
      printf("Dir (len=%x): ", ucLenDI);
      sidDirNew          = simm_malloc(sizeof(_AtapiDirStructType));
      pptDirCur->sidNext = sidDirNew;
      
      pptDirNew = addrC(sidDirNew);
      memset(pptDirNew->szName, 0, sizeof(pptDirNew->szName));
      _AtapiValidateName(&paucSectorBuffer[_BP9], ucLenDI);
      strcpy(pptDirNew->szName, &paucSectorBuffer[_BP9]);
      pptDirNew->ulExtent = _CLong((xdata unsigned char *)&paucSectorBuffer[_BP3]);
      pptDirNew->iPDN     = _CInt((xdata unsigned char *)&paucSectorBuffer[_BP7]);

      printf("(%s)- PDN = %d, Ext = %4lx, Sid = %4lx\r\n", pptDirNew->szName, pptDirNew->iPDN, pptDirNew->ulExtent, sidDirNew);

      // Adjust paucSectorBuffer for length of directory, and check for odd/even length
      paucSectorBuffer += ((_BP9) + ucLenDI);
      if ((ucLenDI & 1) != 0) paucSectorBuffer++;

      pptDirCur = addrC(sidDirNew);
    }
  } // while
  
  m_ucAtapiDirListSize[p_ucDevice] = ucDirCount;
  printf("DirCount = %d\r\n", (int)ucDirCount);
  printf("DirListCreate--\r\n");
}

/*********************
 * AtapiDirListDelete
 *	Public function that deletes the list created by AtapiDirListCreate.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiDirListDelete(unsigned char p_ucDevice) reentrant
{
  static _AtapiDirStructType* pptDirRoot;
  static _AtapiDirStructType* pptDirCur;
  static simm_id              sidCur;
  static simm_id              sidDel;
    
  pptDirRoot         = &m_AtapiDirList[p_ucDevice];
  pptDirCur          = pptDirRoot;
  sidCur             = pptDirCur->sidNext;
  pptDirCur->sidNext = 0;
  
  while (sidCur != 0)
  {
    sidDel = sidCur;
    sidCur = ((_AtapiDirStructType*)addrC(sidDel))->sidNext;
    simm_free(sidDel);
  }
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - File list "class"
//
// ======================================================================

/*********************
 * AtapiFileListIterateInit
 *	Public function that initializes the iterator values before
 *	AtapiFileListIterateNext() can be called.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *
 */
void AtapiFileListIterateInit(unsigned char p_ucDevice)
{
  m_sidAtapiFileListIterator[p_ucDevice] = m_AtapiFileList[p_ucDevice].sidNext;
}

/*********************
 * AtapiFileListIterateNext
 *	Public function that fetches the next file in the FileList for given device.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *
 *	Returns true if a file found, otherwise if list is empty returns false
 *
 */
unsigned char AtapiFileListIterateNext(unsigned char p_ucDevice,
                                       simm_id*      p_psidFileEntry)
{
  simm_id sidFileCur;
  
  sidFileCur = m_sidAtapiFileListIterator[p_ucDevice];
  
  if (sidFileCur == 0)
    return false;
    
  // return data
  *p_psidFileEntry = sidFileCur;
  
  // Point to next in list
  m_sidAtapiFileListIterator[p_ucDevice] = ((_AtapiFileStructType*)(addrC(sidFileCur)))->sidNext;
  return true;
}

/*********************
 * AtapiFileListCreate
 *	Public function that creates the full file list for a single directory.
 *	It traverses the directory to fetch the ATAPI specific file list and puts them
 *	in a module global list, which is iterated later on with the previous functions.
 *
 *	This list will be used by the DirList/PlayList modules to retrieve the data &
 *	build the directory and playlists. This list therefore is a temporary list which
 *	will be deleted after its use.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiFileListCreate(unsigned char p_ucDevice,
                         simm_id       p_sidDirEntry)

{
  static xdata unsigned char        *paucSectorBuffer;
  static xdata unsigned char         ucLenDR;
  static xdata unsigned char         ucLenFI;
  static xdata unsigned char         ucFlags;
  static xdata _AtapiFileStructType* pptFileRoot;
  static xdata _AtapiFileStructType* pptFileCur;
  static xdata simm_id               sidFileNew;
  static xdata _AtapiFileStructType* pptFileNew;
  static volatile unsigned char      ucExtentCount;
  static _AtapiDirStructType        *pDirEntry;
  static xdata unsigned long         ulExtent;
  static xdata unsigned long         ulDataLength;
  
  // Use static var to overcome SLOC space and copy parameter value
  static unsigned char ucDevice;
  
  ucDevice    = p_ucDevice;
  pptFileRoot = &m_AtapiFileList[p_ucDevice];
  pptFileCur  = pptFileRoot;

  // Assume dir is valid dir...
  pDirEntry = (_AtapiDirStructType*)addrC(p_sidDirEntry);
  ulExtent  = pDirEntry->ulExtent;
  printf("ATAPI-FileListCreate: Sid = %4lx, Dir = %s, Extent = %4lx\r\n", p_sidDirEntry, pDirEntry->szName, pDirEntry->ulExtent);
  
  if (ulExtent == 0) return;
  
  do
  {
    AtapiReadSector2(ucDevice, //p_ucDevice,
                     ulExtent);
    ulExtent++;
             
    // Size is also stored in directory record. We must read this record, in order to know the size,
    // and thus the number of sectors to read...
    // Always read & interpret the first sector. Check for each dir if it's the "current" directory which
    // names consists of a single (00) byte, extract size, and calculate #of sectors to read...
  
              
    // Format of Directory Record:
    // ----------------------------
    // _BP			Field Name					Content
    // ==			==========					=======
    // 1			Length of Directory Record (LEN_DR)		numerical value
    // 2			Extended Attribute Record Length		numerical value
    // 3 to 10			Location of Extent				numerical value
    // 11 to 18			Data Length					numerical value
    // 19 to 25			Recording Date and Time				numerical values
    // 26			File Flags					8 bits
    // 27			File Unit Size					numerical value
    // 28			Interleave Gap Size				numerical value
    // 29 to 32			Volume sequence Number				numerical value
    // 33			Lenght of File Identifier (LEN_FI)		numerical value
    // 34 to (33 + LEN_FI)	File Identifier					d-chars, d1-chars, SEP1, SEP2, (00) or (O1) byte
    // 34 + LEN_FI		Padding Field					(00) byte
    // (LEN_DR - LEN_SU + 1)	System Use					LEN_SU bytes
    //   to LEN_DR
  
    paucSectorBuffer = m_aucSectorBuffer;
    
    // Get size of directory record
    // A size of 0 marks the end of the records...
    
    ucLenDR = paucSectorBuffer[_BP1];
//    printf("Size of 1st dir = %x", ucLenDR);
    while (ucLenDR) //(0 != ucLenDR)
    {
      ucLenFI = paucSectorBuffer[_BP33];
//      printf("(ucLenFI = %d)", (int)ucLenFI);
      ulDataLength = *((unsigned long *)(paucSectorBuffer+_BP11));

      ucFlags = (paucSectorBuffer[_BP26]);
//      printf(", flags = %x\r\n", ucFlags);
      
      // Check if it's a file, not a directory...
      if ((ucFlags & _BP26_Directory) == 0x0)
      {
        sidFileNew          = simm_malloc(sizeof(_AtapiFileStructType));
        pptFileCur->sidNext = sidFileNew;
        pptFileNew          = addrC(sidFileNew);
        
        memset(pptFileNew->szName, 0, sizeof(pptFileNew->szName));
        memcpy(pptFileNew->szName, &paucSectorBuffer[_BP34], ucLenFI);
        _AtapiValidateName(pptFileNew->szName, ucLenFI);
        pptFileNew->sidNext      = 0;
        pptFileNew->ulDataLength = ulDataLength;
        pptFileNew->ulExtent     = _CLong((xdata unsigned char *)&paucSectorBuffer[_BP3]);
        pptFileCur               = pptFileNew;
        printf("-Found: <%s>, extent = %lx, length = %lx\r\n", pptFileNew->szName, pptFileNew->ulExtent, pptFileNew->ulDataLength);
      }
      else
      {
      	if ((ucLenFI == 1) && ((unsigned char)paucSectorBuffer[34] == '\0'))
      	{
      	  // This is the "Current" directory record. Calculate # of sectors this extent is.
      	  // Calculation is simple. Size is given in bytes rounded to the sector size (2048 bytes).
          ucExtentCount = (unsigned char)(unsigned long)(ulDataLength / 2048L);
      	}
      }
      // The next line does NOT work. Don't ask me why. It works fine in other functions...
      //   paucSectorBuffer += (0 + ucLenDR);
      // So replaced by very efficient loop...
      while (ucLenDR-- > 0) paucSectorBuffer++;
      ucLenDR = paucSectorBuffer[_BP1];
    }
  } while (--ucExtentCount > 0);
}

/*********************
 * AtapiFileListDelete
 *	Public function that deletes the list created by AtapiFileListCreate.
 *
 *	I could draw some nice UML diagrams, and throw in some design patterns, but I don't!
 *	People with X and/or OSF/Motif knowledge should know this approach...
 *
 */
void AtapiFileListDelete(unsigned char p_ucDevice) reentrant
{
  static _AtapiFileStructType* pptFileRoot;
  static _AtapiFileStructType* pptFileCur;
  static simm_id               sidCur;
  static simm_id               sidDel;
    
  pptFileRoot = &m_AtapiFileList[p_ucDevice];
  pptFileCur  = pptFileRoot;
  sidCur      = pptFileCur->sidNext;
  while (sidCur != 0)
  {
    sidDel = sidCur;
    sidCur = ((_AtapiFileStructType*)addrC(sidDel))->sidNext;
    simm_free(sidDel);
  }
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - ATAPI I/O functions
//
// ======================================================================

/*********************
 * AtapiVolumeRecognitionSequence
 *	Public function that performs the basic strategy of CD-ROM volume recognition,
 *	which is a sequence of volume descriptors, recorded one per (logical) sector, 
 *	starting at Sector 16 in the first track of the last session on the disc.
 *
 *	The starting track of the last session on the disc must be fetched via the
 *	TOC (Table Of Contents). THIS HAS NOT YET BEEN IMPLEMENTED, meaning that
 *	only those songs in session 1 on the disc will be recognized...
 *
 *	Currently only the following volumes & filesystem types are recognized:
 *	- PVD with filesystem DOS, ie. the 8.3 file format
 *	- SVD with filesystem Joliet, ie. Microsofts long file name format with
 *	  UCS-2 Level 1, level 2 and level 3 formats.
 *
 *	So, not supported are:
 *	- SVD with filesystems Romeo and Rockridge!!
 *
 *	Returns true if volume is recognized, otherwise false.
 *	In some cases a retry seems to be necessary to recognize the volume correctly.
 *	This is something for later to find out why this sometimes happens !
 *
 */
unsigned char AtapiVolumeRecognitionSequence(unsigned char p_ucDevice,
                                             unsigned long p_ulStartLSN)

{
  unsigned char                       ucVolumeDescr;
  unsigned long                       ulSectorCur;
  _AtapiVolumeDescriptorStructTypePtr pCurVds;
  
  // Initialize vds for this device
  memset(m_vds[p_ucDevice], 0, sizeof(_AtapiVolumeDescriptorSet));
  
  ulSectorCur = p_ulStartLSN;
  
  do
  {
    AtapiReadSector2(p_ucDevice, ulSectorCur);
    //_AtapiDebugPrintSectorBuffer();

    // Copy volume descriptor. SDCC seems to f**k up somewhere, resulting in terminating
    // the sequence almost immediately, instead of locating the volumes on the CD...
    ucVolumeDescr = *m_aucSectorBuffer;

    // Check signature (Standard Identifier). Return if not "CD001".
    if (*(m_aucSectorBuffer + 1)!= 'C' ||
	*(m_aucSectorBuffer + 2)!= 'D' ||
	*(m_aucSectorBuffer + 3)!= '0' ||
	*(m_aucSectorBuffer + 4)!= '0' ||
	*(m_aucSectorBuffer + 5)!= '1'   )
    {
      printf("Volume NOT recognized, signature CD001 not found\r\n");
      return(false);
    }

    switch (ucVolumeDescr) //((unsigned char)(*m_aucSectorBuffer)))
    {
      // Boot Volume
      case _ucVOLUME_DESCR_TYPE_BOOT :
      {
         printf("BOOT...\r\n");
	 //memcpym_aucSectorBuffer+7,pCurVds->BootSystemIdentifier,32);
	 //memcpym_aucSectorBuffer+39,pCurVds->BootIdentifier,32);
	 break;
      }
      
      // Primary Volume Descriptor
      case _ucVOLUME_DESCR_TYPE_PVD :
      {
        printf("PVD...\r\n");
        pCurVds = &(m_vds[p_ucDevice][_VDSPVD]);

        _AtapiGetVdsData(pCurVds, m_aucSectorBuffer);
        
        // The PVD is always considered as having _ucFSF_DOS as the file system format.
        // Make sure that the Vds data is correctly filled for this variable.
	pCurVds->ucFileSystemFormat = _ucFSF_DOS;
	break;
      }
      
      // Secondary Volume Descriptor
      case _ucVOLUME_DESCR_TYPE_SVD :
      {
        printf("SVD...\r\n");

        // 8.5.3 Volume Flags (_BP 8)
        // 
        // The bits of this field shall be numbered from O to 7 starting with the least significant bit.
        //
        // This field shall specify certain characteristics of the volume as follows.
        // Bit O: - if set to ZERO, shall mean that the Escape Sequences field specifies
        //          only escape sequences registered according to ISO 2375;
        //        - if set to ONE, shall mean that the Escape Sequences field specifies
        //          at least one escape sequence not registered according to ISO 2375.
        //
        // Bits 1 to 7: These bits are reserved for future standardization and shall all be set to ZERO.

        // The UCS-2 Level 1, UCS Level 2, and UCS-2 Level 3 escape sequences are considered to be registered
        // according ISO 2735 for purposes of setting bit 0 of the Volume Flags field of the SVD to ZERO
        if ((*(m_aucSectorBuffer + _BP8) & 0x01) == 0)
	{
          // Check escape sequence for Microsofts ISO/Joliet formats...
	  if ((*(m_aucSectorBuffer + _BP89) == 0x25) &&
	      (*(m_aucSectorBuffer + _BP90) == 0x2F))
	  {
            if ((*(m_aucSectorBuffer + _BP91) == _ucFSF_JOLIET_LVL1) ||
		(*(m_aucSectorBuffer + _BP91) == _ucFSF_JOLIET_LVL2) ||
		(*(m_aucSectorBuffer + _BP91) == _ucFSF_JOLIET_LVL3))
            {
              printf("... Joliet\r\n");
              pCurVds = &m_vds[p_ucDevice][_VDSSVD];
	      _AtapiGetVdsData(pCurVds, m_aucSectorBuffer);
            }
          }
	}
	break;
      }
      
      // Volume Partition Descriptor
      case _ucVOLUME_DESCR_TYPE_VPD :
      {
        break;
      }
      
      // Volume delimiter. This is the end...
      case _ucVOLUME_DESCR_TYPE_DELIMITER :
      {
        break;
      }
      
      default :
      {
        return(false);
      }
    }
    
    ulSectorCur++;
  }
  while (ucVolumeDescr != _ucVOLUME_DESCR_TYPE_DELIMITER);
  //((unsigned char)*m_aucSectorBuffer)!= _ucVOLUME_DESCR_TYPE_DELIMITER);

  return(true);
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - ATAPI commands
//
// ======================================================================

/*********************
 * AtapiRequestSense
 *	Public function that senses the last error information from the drive.
 *
 */
char AtapiRequestSense(unsigned char p_ucDevice)
{
  p_ucDevice; // Prevent compiler warning. Remove when code below is made active again
// def out. Need the code space !!!!!!!!!!
/*
  xdata AtapiPacketBufferType aucCdb;
  char                        bResult = false;
  int                         iByteCount;
  
  printf("ATAPI-RequestSense\r\n" );

  memset(m_aucCommandBuffer, 0, sizeof(m_aucCommandBuffer));
  memset(aucCdb, 0, sizeof(aucCdb));
  
  aucCdb[0] = cmdATAPI_REQUEST_SENSE;	// Request sense command
  aucCdb[4] = 32;			// Buffer allocation length.
  
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);
  
  if (!bResult)
  {
    printf("ATAPI-RequestSense: ERROR!\r\n");
  }
  else
  {
    printf( "Error code= %x", (char)m_aucCommandBuffer[0]);
    printf( ", Sense Key=%x", ((char)(m_aucCommandBuffer[2] & 0x0f)));
    printf( ", ASC=%x", (char)m_aucCommandBuffer[12]);
    printf( ", ASCQ=%x", (char)m_aucCommandBuffer[13]);
    printf("\n");
  }

  return bResult;
*/
}

/*********************
 * AtapiEject
 *	Public function that ejects the disc from the drive...
 *
 */
char AtapiEject(unsigned char p_ucDevice)

{
  xdata AtapiPacketBufferType aucCdb;
  char                        bResult = false;
    
  printf("ATAPI-Eject\r\n" );
  memset(aucCdb, 0, sizeof(aucCdb));
  aucCdb[0] = cmdATAPI_START_STOP;   // StartStop command code
  aucCdb[4] = 2;      // params: Eject
  memset(m_aucCommandBuffer, 0, sizeof(m_aucCommandBuffer) );
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);

  if (!bResult)
    printf("ATAP-Eject: ERROR!\r\n");

  return bResult;
}

/*********************
 * AtapiStartStop
 *	Public function that performs starting and stopping the drive.
 *	Start works, but the stop is ignored for this moment, which is not a real problem since
 *	the drive will stop spinning if no read requests are issued anymore...
 *
 */
char AtapiStartStop(unsigned char p_ucDevice, char bStart)
{
  xdata AtapiPacketBufferType aucCdb;
  char                        bResult = false;
  unsigned char               ucStatus;
  
  bStart; // Prevent compiler warning
  
  printf("ATAPI-StartStop\r\n" );
  memset(aucCdb, 0, sizeof(aucCdb));
  aucCdb[0] = cmdATAPI_START_STOP;   // StartStop command code
  aucCdb[4] = 1;      // params: Start
  memset(m_aucCommandBuffer, 0, sizeof(m_aucCommandBuffer) );
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);

  // Drive has been started. Now wait for it to become ready.
  // This is done by sending a NULL command to the drive, as long as the drive's sensekey returns "2".
  
  memset(aucCdb, 0, sizeof(aucCdb));
  while (1)
  {
    bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);

    ucStatus = AtaReadRegister(regATA_ALTSTAT);
    printf("AS = %x, ", ucStatus);
    ucStatus = AtaReadRegister(regATA_ERROR);
    printf("E = %x, ", ucStatus);
    ucStatus = (ucStatus & mskATAPI_ER_SNSKEY);
    ucStatus = ucStatus >> 4;
    printf("SnsKey = %x\r\n", ucStatus);
    AtapiRequestSense(p_ucDevice);

    if (ucStatus != 2)
      break;
    // There should be a 100msec delay between the commands...
    // My DVD drive goes wild, if I let this delay out !?!?!
    _AtapiDelay100ms();    
  }
  
  return 0;
}

/*********************
 * AtapiReadTOC
 *	Public function that reads the TOC (Table of Contents) stored on the CD.
 *	This gives information about the number of sessions recorded, and their starting
 *	and ending/length positions.
 *	We should fetch the extent of the last recorded session to read the latest
 *	volume descriptors, but by this time that is not yet implemented.
 *
 */
char AtapiReadTOC(unsigned char p_ucDevice)
{
  xdata AtapiPacketBufferType aucCdb;
  char                     bResult = false;
  int                      iByteCount;
  
  printf("ATAPI-ReadTOC\r\n" );

  memset(m_aucCommandBuffer, 0, sizeof(m_aucCommandBuffer));
  memset(aucCdb, 0, sizeof(aucCdb));
  
  aucCdb[0] = cmdATAPI_READ_TOC;// Read TOC command
  aucCdb[1] = 0x00;		// MSF Flag to zero (we want LBA addressing)
  aucCdb[2] = 0x02;		// Full TOC
  aucCdb[7] = 0x02;		// Allocation length...
  aucCdb[8] = 0x00;		// ...of 512 bytes
  
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);
  
  if (!bResult)
  {
    printf("ATAP-ReadTOC: ERROR!\r\n");
  }
  else
  {
    iByteCount = (int)m_aucCommandBuffer[0]<<8 | (int)m_aucCommandBuffer[1];
    printf("Bytes2xfer = %d\r\n", iByteCount);
    printf( "1st Session= %x", m_aucCommandBuffer[2]);
    printf(", last Session= %x\r\n", m_aucCommandBuffer[3]);
    // TODO:
    // Determine extent of last session, and use this as the basis to read the PVD etc.
    // m_tDeviceProperties[p_ucDevice].ulExtentStart = <extent of last session on disk>;
  }

  return bResult;
}

/*********************
 * AtapiReadCapacity(
 *	Public function that reads the capacity of the disk.
 *	This is not only a nice to have function. The maximum extent can be used to
 *	verify that read requests do NOT read beyond the disksize.
 *
 */
char AtapiReadCapacity(unsigned char p_ucDevice)
{
  xdata AtapiPacketBufferType aucCdb;
  char                     bResult = false;
    
  printf("ATAPI-ReadCapacity\r\n" );
  
  memset(m_aucCommandBuffer, 0, sizeof(m_aucCommandBuffer));
  memset(aucCdb, 0, sizeof(aucCdb));

  aucCdb[0] = cmdATAPI_READ_CAPACITY;	// Read capacity command
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), &aucCdb, sizeof(m_aucCommandBuffer), &m_aucCommandBuffer);

  if (!bResult)
  {
    printf("ATAP-ReadCapacity: ERROR!\r\n");
  }
  else
  {
    // Get capacity and store this one. It can be used to check for valid extents...
    m_tDeviceProperties[p_ucDevice].ulExtentMax = _CLong(&m_aucCommandBuffer[0]);
    printf("-Capacity (extents): %lx\r\n", m_tDeviceProperties[p_ucDevice].ulExtentMax);

    // Validate that size of a single extent is indeed 2048 bytes...
    if (!(m_aucCommandBuffer[4] == 0x00 && m_aucCommandBuffer[5] == 0 &&
          m_aucCommandBuffer[6] == 0x08 && m_aucCommandBuffer[7] == 0))
    {
      printf("-!!! Blocklength != 2048 !!!\r\n");
      bResult = false;
    }
  }
  
  return bResult;
}

/*********************
 * AtapiReadSector2
 *	Reads the given extent from the given device.
 *	Size of data to read is always one extent, ie a block of 2KB.
 *
 */
char AtapiReadSector2(unsigned char p_ucDevice,
                      unsigned long p_ulExtent)
{
  static unsigned char aucLBA[3];
  
  _Long2Char(p_ulExtent, &aucLBA);
  return AtapiReadSector(p_ucDevice, aucLBA[0], aucLBA[1], aucLBA[2], aucLBA[3]);
}
/*********************
 * AtapiReadSector
 *	Reads the given extent from the given device.
 *	Size of data to read is always one extent, ie a block of 2KB.
 *
 */
char AtapiReadSector(unsigned char p_ucDevice,
                     unsigned char p_ucLBA0,
                     unsigned char p_ucLBA1,
                     unsigned char p_ucLBA2,
                     unsigned char p_ucLBA3)
{
  xdata AtapiPacketBufferType aucCdb;
  char  bResult = false;
    
  printf("ATAPI-Read: %1x-%1x-%1x-%1x\r\n", (char)p_ucLBA3, (char)p_ucLBA2, (char)p_ucLBA1, (char)p_ucLBA0);
  memset(aucCdb, 0, sizeof(aucCdb));
  aucCdb[0] = 0xa8;		// read command code
  aucCdb[2] = p_ucLBA3;
  aucCdb[3] = p_ucLBA2;
  aucCdb[4] = p_ucLBA1;
  aucCdb[5] = p_ucLBA0;
  aucCdb[9] = 1;		// # of sectors to read (1 by default !!)

  memset(m_aucSectorBuffer, 0, sizeof(m_aucSectorBuffer));
  bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), aucCdb, sizeof(m_aucSectorBuffer), &m_aucSectorBuffer);

  if (!bResult)
    printf("ATAP-Read: ERROR!\r\n");

  return bResult;
}

/*********************
 * AtapiSendCommand
 *	Writes a command to the ATA command registers, followed by the ATAPI command buffer
 *
 * NOTE:
 *	This VERY LONG routine is likely to be rewritten in assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 */
char AtapiSendCommand(unsigned char  p_ucDev,
                      unsigned int   p_uiCdbSize,
                      unsigned char* p_paucCdb,
                      unsigned int   p_uiBufferSize,
                      unsigned char* p_paucBuffer)
{
    unsigned char ucRegDevCtrl;
    unsigned char ucRegDevHead;
    unsigned char ucRegCylLo;
    unsigned char ucRegCylHi;
    unsigned char ucRegFr;
    unsigned char ucRegSc;
    unsigned char ucRegSn;
    unsigned char ucStatus;
    unsigned char ucRegReason;
    unsigned char ucRegLowCyl;
    unsigned char ucRegHighCyl;
    unsigned int  uiByteCnt;
    unsigned int  uiWordCnt;
    unsigned char ucXferData;
    char          bResult;
   
   // Assume no errors found...
   bResult = true;

   // Assume the command will read data from the ATAPI device...
   ucXferData = _iXFER_READ;
   
   // setup register values
   ucRegFr = ucRegSc = ucRegSn = 0;

   // Setup device control register
   ucRegDevCtrl = regATA_DEVCTRL_HD15 | regATA_DEVCTRL_NIEN;
   // Select the pri or sec device
   ucRegDevHead = ( p_ucDev ? regATA_DEVHEAD_DEV1 : regATA_DEVHEAD_DEV0 ) | 0x0f;
   // Determine the number of bytes (maximum) to transfer from/to atapi device from passed buffersize
   // and put them in cylLow and cylHigh...
   ucRegCylLo =  p_uiBufferSize & 0x00ff;
   ucRegCylHi = (p_uiBufferSize & 0xff00) >> 8;

   AtaSelectDevice(p_ucDev);

   // Set up all the registers except the command register.
   AtaWriteRegister(regATA_DEVCTRL, ucRegDevCtrl);
   AtaWriteRegister(regATA_FEATURE, ucRegFr);
   AtaWriteRegister(regATA_SECCNT,  ucRegSc);
   AtaWriteRegister(regATA_SECNR,   ucRegSn);
   AtaWriteRegister(regATA_CYLMSB,  ucRegCylHi);
   AtaWriteRegister(regATA_CYLLSB,  ucRegCylLo);
   AtaWriteRegister(regATA_DEVHEAD, ucRegDevHead);

   // Start the command by setting the Command register.  The drive
   // should immediately set BUSY status.

   AtaWriteRegister(regATA_DEVCMD, cmdATA_PACKET);

   // Check for protocol failures & wait until busy is set (endless loop...)

   ucStatus = AtaReadRegister(regATA_ALTSTAT);
   while (1)
   {
      if (ucStatus & regATA_PRISTAT_BSY)
         break;                              // BSY=1 (that's OK)
      else
      {
         if (ucStatus & regATA_PRISTAT_DRQ)
            break;                           // BSY=0, DRQ=1 (that's OK)
         if (ucStatus & regATA_PRISTAT_ERR)
            break;                           // BSY=0, error (that's OK)
      }
      ucStatus = AtaReadRegister(regATA_ALTSTAT);
   }

   // Poll Alternate Status for BSY=0.

   while (1)
   {
      ucStatus = AtaReadRegister(regATA_ALTSTAT);       // poll for not busy
      if ((ucStatus & regATA_PRISTAT_BSY) == 0)
         break;
   }

   if (bResult == true)
   {
      // Read the primary ucStatus register and the other ATAPI registers.

      ucStatus     = AtaReadRegister(regATA_PRISTAT);
      ucRegReason  = AtaReadRegister(regATA_SECCNT);
      ucRegLowCyl  = AtaReadRegister(regATA_CYLLSB);
      ucRegHighCyl = AtaReadRegister(regATA_CYLMSB);

      // check status: must have BSY=0, DRQ=1 now

      if ((ucStatus & (regATA_PRISTAT_BSY | regATA_PRISTAT_DRQ)) != regATA_PRISTAT_DRQ)
      {
         printf("-DRQ not set...\r\n");
         bResult    = false;
         ucXferData = _iXFER_NONE;
      }
      else
      {
         // Do some sanity checks. Should not be necessary, since it is ignored...
         // check: C/nD=1 == COMMAND, IO=0 (== WRITE).

         if ((ucRegReason & (regATA_SECCNT_P_TAG | regATA_SECCNT_P_REL | regATA_SECCNT_P_IO)) ||
             (!(ucRegReason &  regATA_SECCNT_P_CD)))
           printf("ATAPI- protocol command or I/O failure\r\n");
           
         // Check size returned...
         if ((ucRegLowCyl != ucRegCylLo) || (ucRegHighCyl != ucRegCylHi))
           printf("ATAPI - sizes dif\r\n");

         AtapiWriteData(p_paucCdb, p_uiCdbSize >> 1 );
      }
   }

   // Data transfer loop...
   // Note that a single request can be split up by the drive into multiple answers...
   // That's why it is a loop, and not a single line...

   while (true == bResult)
   {
      // We should have a 100 msec delay build into this part following ATAPI specs,
      // but we don't. The software seems to be slow enough not having a extra delay build in...
      HOST_BSY0;

      // Data transfer loop...
      // If there was a time out error, exit the data transfer loop.

      if (false == bResult)
      {
          printf("-errors, break\r\n");
          ucXferData = _iXFER_NONE;
          break;
      }

      // Data transfer loop...
      // Read the primary status register and the number of bytes the drive wants to transfer

      ucStatus     = AtaReadRegister(regATA_PRISTAT);
      ucRegReason  = AtaReadRegister(regATA_SECCNT);
      ucRegLowCyl  = AtaReadRegister(regATA_CYLLSB);
      ucRegHighCyl = AtaReadRegister(regATA_CYLMSB);

      // Data transfer loop...
      // Exit the read data loop if the device indicates this is the end of the command.

      if ((ucStatus & (regATA_PRISTAT_BSY | regATA_PRISTAT_DRQ)) == 0)
      {
          printf("-cmd done\r\n");
          ucXferData = _iXFER_NONE;
          break;
      }

      // Data transfer loop...
      // The device must want to transfer data...
      // check status: must have BSY=0, DRQ=1 now.
      // This part sometimes gives errors which I can not understand...

      if ((ucStatus & (regATA_PRISTAT_BSY | regATA_PRISTAT_DRQ)) != regATA_PRISTAT_DRQ)
      {
          printf("-E AtapiSendCommand: ucStatus !\r\n");
          bResult    = false;
          ucXferData = _iXFER_NONE;
          break;
      }
      else
      {
        // Get nr of bytes to be transferred by the device.
        // I had some strange values read back in the beginning of this implementation,
        // the drive wants to transfer data, but tells me also it's about 0 bytes...
        
        uiByteCnt = ((int)ucRegHighCyl << 8) | ucRegLowCyl;
        if (uiByteCnt < 1)
        {
           printf("ATAPI- byteCnt < 1\r\n");
           bResult    = false;
           ucXferData = _iXFER_NONE;
           break;
        }

        // The number of bytes requested should match the number of bytes the drive
        // is reporting to transfer. Make sure this is the case. Again, in the beginning of
        // this implementation, wierd things occurred (they stopped however...)

        if ((uiByteCnt) > p_uiBufferSize)
        {
          printf("ATAPI- transfer (%x) > maxbytes (%x)\r\n", uiByteCnt, p_uiBufferSize);
          bResult    = false;
          ucXferData = _iXFER_NONE;
          break;
        }

        // Actually transfer the data to or from the drive.
        // Data transfer is always done in words, not in bytes, so convert & adjust odd values.

        uiWordCnt = (uiByteCnt >> 1) + (uiByteCnt & 0x0001);

        if (ucXferData == _iXFER_WRITE)
        {
      	  printf("Writing %d words to ATAPI...\r\n", uiWordCnt);
          AtapiWriteData(p_paucBuffer, uiWordCnt);
        }
        else
        // ucXferData == _iXFER_READ !!!
        {
      	  printf("ATAPI-Read %d words\r\n", uiWordCnt);
          AtapiReadData(p_paucBuffer, uiWordCnt);
          
          // Add bytes read to bufferpointer
          // The drive may have split the single data request into multiple data transfers.
          // This happens at least if more than 1 sector (2048 bytes) is requested, in that case the data
          // will be transferred sector by sector...
          p_paucBuffer += uiByteCnt;
        }
     }
  }
  
  if ((true == bResult) && (ucXferData != _iXFER_NONE))
  {
    // We should add a 100msec delay in here following ATAPI spec...
    HOST_BSY0;
  }

  return bResult;
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - ATAPI Low level I/O (PIO)
//
// ======================================================================


/*********************
 * AtapiWriteData
 *	Writes a number of words to the selected ATAPI device using PIO mode.
 *
 *	The data is written from the passed buffer. No size checks are done. It is assumed
 *	the caller supplies a large enough buffer to hold the data to be written
 *
 * NOTE1:
 *	This routine is likely to be rewritten in assembly. For debugging purposes however
 *	(and speed is not an issue yet), this routine is used...
 * NOTE2:
 *	The only data that gets written is normally a 6-word ATAPI command, not more, not less
 *	So IF speed becomes an issue, don't bother about these 12 bytes...
 *
 */
void AtapiWriteData(unsigned char* p_paucData,
                    unsigned char  p_ucWordCnt)
{
  unsigned char i;
  static unsigned char ucHi, ucLo;

#ifdef ATAPI_DEBUG
  printf("ATAPI-WriteData: ");
#endif // ATAPI_DEBUG
  for (i=0 ; i < p_ucWordCnt; i++)
  {
    ucLo = *(p_paucData++);
    ucHi = *(p_paucData++);
    ide_write_dataregisters(ucHi, ucLo);
#ifdef ATAPI_DEBUG
    print_hex16(ucLo);
    printf("-");
    print_hex16(ucHi);
    printf("-");
    //printf("%2x-%2x-", ucLo, ucHi);
#endif // ATAPI_DEBUG
  }
#ifdef ATAPI_DEBUG
  printf("\r\n");
#endif // ATAPI_DEBUG
}

/*********************
 * AtapiReadData
 *	Reads a number of words from the selected ATAPI device using PIO mode.
 *
 *	The data is read into the passed buffer. No size checks are done. It is assumed
 *	the caller supplies a large enough buffer to hold the requested data.
 *
 * NOTE1:
 *	This routine has already been rewritten in assembly. Not only for the sake of speed,
 *	but because of compiler f*** ups. I could NOT get it to work reliable, so had to
 *	rewrite the stuff in assembly, which was done in a snap...
 * NOTE2:
 *	The measured throughput of this piece of software is about 40KB/sec. Its primary use
 *	therefore will be for the (simple) CD recognition requests, not for transferring
 *	the actual song data, this will be done using DMA (either Paul will program this,
 *	or he will explain me how to do that. It will look so simple I guess?!?!?!?!).
 */
void AtapiReadData(unsigned char* p_paucData,
                   unsigned int   p_uiWordCnt)
{
  unsigned char* paucData;
  
  // Prevent the compiler from removing the parameter...
  paucData    = &p_paucData;
  // Prevent compiler warning unrefernced argument
  p_uiWordCnt;
  
//  paucData = p_paucData;
  printf("ATAPI-ReadData (%x)\r\n", p_paucData);
  
/*
  unsigned char  ucHi, ucLo;
  unsigned int   uiData;
  unsigned char* paucData;
  unsigned int   uiIdx;
  
#ifdef ATAPI_DEBUG
  unsigned int  uiCounter = 0;
#endif // ATAPI_DEBUG
  
#ifdef ATAPI_DEBUG
  printf("Read:\r\n ");
  printf("p_paucData = %x\r\n", p_paucData);
  printf("m_aucSectorBuffer = %x\r\n", m_aucSectorBuffer);
#endif // ATAPI_DEBUG
  paucData = p_paucData;
  uiIdx    = 0;
  
  for (; p_uiWordCnt > 0; p_uiWordCnt--)
  {
  // Note that the #else part is faster, but I needed the first part
  // for debugging purposes...
#if 1
    uiData        = ide_read_dataregisters();
    ucLo          = (uiData & 0xff);
    ucHi          = (uiData >> 8);
//    *(paucData++) = ucLo;
//    *(paucData++) = ucHi;
    paucData[uiIdx]   = ucLo;
    paucData[uiIdx+1] = ucHi;
    uiIdx             += 2;
#else
//  DOES NOT WORK PROPERLY !!!!!!!!!!!
    *(paucData++) = ucLo = reg_ide_data;
    *(paucData++) = ucHi = reg_ide_data_msb;
#endif
    
#ifdef ATAPI_DEBUG
    // Do the printing/debugging stuff
    uiCounter++;
    // print_hex16 seems to flip hi/lo bytes. So swap again...
    uiData = (((unsigned int)(ucLo)) << 8) | (unsigned int)ucHi;
    print_hex16(uiData);
    if ((uiCounter % 10) == 0) print_crlf();
#endif // ATAPI_DEBUG
  }
*/
	_asm
;	Get pointer to buffer to write data into
;	NOTE that buffer MUST be referenced in C part, otherwise compiler removes it (optimized)	
	mov	dptr,#_AtapiReadData_p_paucData_1_1
	movx	a,@dptr
	mov	r4,a
	inc	dptr
	movx	a,@dptr
	mov	r5,a
	mov	dpl,r4
	mov	dph,r5
;
;	Save pointer to buffer
	push	dph
	push	dpl
;
#ifdef ATAPI_DEBUG
;	Print value of buffer to store data
	lcall	_print_hex16
	lcall	_print_crlf
#endif
;
;	Get number of words to read
	mov	dptr,#_AtapiReadData_PARM_2
	movx	a,@dptr
	mov	r4,a
	inc	dptr
	movx	a,@dptr
	mov	r5,a
;
#ifdef ATAPI_DEBUG
;	Print value of #words to read
	mov	dpl,r4
	mov	dph,r5
	lcall	_print_hex16
	lcall	_print_crlf
#endif
;	
;	Restore pointer to buffer	
	pop	dpl
	pop	dph
;
rdblk2: push    dph
        push    dpl
;
;	Next 2 lines do something special. I do not exactly know the consequences
;	of these instructions, but if left out, data becomes garbled.
;	(Paul will have an answer ?)
	mov	c,ea
	clr	ea
        mov     dptr, #_ucregATA_DATALSB
        movx    a,@dptr			; read lsb
        mov     r2,a
        mov     dptr, #_ucregATA_DATAMSB
        movx    a,@dptr			; read msb
;
;	Next line does something special. I do not exactly know the consequence
;	of this instruction, but if left out, data becomes garbled.
;	(Paul will have an answer ?)
	mov	ea,c
        mov     r3,a
;
#ifdef ATAPI_DEBUG
        mov	dpl,a			; Is msb
        mov	a,r2			; Is lsb
        mov	dph,a
        lcall	_print_hex16
#endif
;
        pop     dpl
        pop     dph
        mov     a,r2
        movx    @dptr,a			; write lsb
        inc     dptr
        mov     a,r3
        movx    @dptr,a			; write msb
        inc     dptr
;
;	Decrement 16-bit counter, and loop if not yet zero
        dec	r4
        cjne	r4,#0x00,rdblk3
        cjne	r5,#0x00,rdblk3
        sjmp	rdblk4
rdblk3: cjne	r4,#0xff,rdblk2
	dec	r5
        sjmp	rdblk2
rdblk4:
        _endasm;
}

// ======================================================================
//
// ========== PUBLIC FUNCTIONS - File Cache, Uncache functions
//
// ======================================================================

typedef struct
{
  unsigned int blkMemoryBlock;
  simm_id      sidNext;
} AtapiFileCacheEntryType, *AtapiFileCacheEntryTypePtr;
typedef simm_id AtapiFileCacheEntryTypeSid;

static simm_id m_sidFileCacheList[ATA_DEVICE_COUNT_MAX] = {0, 0};
static simm_id m_sidFileCacheCur[ATA_DEVICE_COUNT_MAX] = {0, 0};

void AtapiFileReadInit(unsigned char p_ucDevice)
{
  m_sidFileCacheCur[p_ucDevice] = 0;
}

/*********************
 * AtapiFileReadBlock
 *	Returns the next block in line to be passed (eventually) to the MP3 decoder.
 *	The file/song is previously read into these 4K memory blocks by FileCache and
 *	will be freed by FileUncache.
 *
 * NOTE:
 *	This is just temporary software, to show that ATAPI can play files...
 *
 */
unsigned int AtapiFileReadBlock(unsigned char p_ucDevice)
{
  xdata AtapiFileCacheEntryTypeSid sidEntry;
  xdata AtapiFileCacheEntryTypeSid sidFileCacheCur;
  xdata unsigned int blk;
  
  if (m_sidFileCacheList[p_ucDevice] == 0)
    return (0);

  sidFileCacheCur = m_sidFileCacheCur[p_ucDevice];
  printf("Cur = "); print_hex32(sidFileCacheCur); print_crlf();
  
  if ((unsigned long)sidFileCacheCur == 0xFFFFFFFF)
    return (0xFFFF);
  
  // Get current entry to return in list.
  // Check if this is the first call, and set up iterator, otherwise get next entry
  if (sidFileCacheCur == 0)
  {
    sidFileCacheCur = m_sidFileCacheList[p_ucDevice];
    printf("was 0, now "); print_hex32(sidFileCacheCur); print_crlf();
  }
  else
  {
    sidEntry = ((AtapiFileCacheEntryTypePtr)(addrC(sidFileCacheCur)))->sidNext;
    printf("was "); print_hex32(sidFileCacheCur); printf(", now "); print_hex32(sidEntry); print_crlf();
    sidFileCacheCur = sidEntry;
  }
  
  // If end of list, set to 0xFFFF, meaning end of list...
  if (sidFileCacheCur == 0)
  {
    sidFileCacheCur = 0xFFFFFFFF;
    blk             = 0xFFFF;
    printf("EOL\r\n");
  }
  else
  {
    blk = ((AtapiFileCacheEntryTypePtr)(addrC(sidFileCacheCur)))->blkMemoryBlock;
    printf("RdBlk = %x\r\n", blk);
  }
    
  // Remember where we were in the list...
  m_sidFileCacheCur[p_ucDevice] = sidFileCacheCur;

  return (blk);
}

/*********************
 * AtapiFileUncache
 *	Unreads the file into memory. In other words, memory is freed again...
 *
 */
char AtapiFileUncache(unsigned char p_ucDevice)
{
  xdata AtapiFileCacheEntryTypeSid sidEntry;
  xdata AtapiFileCacheEntryTypeSid sidFileCacheCur;

  printf("ATAPI-Uncache\r\n");
  if (m_sidFileCacheList[p_ucDevice] == 0)
    return (false);
    
  // Cache contains something, free it...
  sidFileCacheCur = m_sidFileCacheList[p_ucDevice];
  
  while (sidFileCacheCur != 0)
  {
    sidEntry = ((AtapiFileCacheEntryTypePtr)(addrC(sidFileCacheCur)))->sidNext;
    free_blocks(((AtapiFileCacheEntryTypePtr)(addrC(sidFileCacheCur)))->blkMemoryBlock);
    print_hex32(sidFileCacheCur); printf(", ");
    simm_free(sidFileCacheCur);
    sidFileCacheCur = sidEntry;
  }
  
  m_sidFileCacheList[p_ucDevice] = 0;
  
  print_crlf();
  
  return true;
}
/*********************
 * AtapiFileCache
 *	Reads the file into memory.
 *	For testing sake (I've got mp3's of 30MB and more...), length is set to a max of 6MB
 *
 * NOTE:
 *	This is just a demo function to proof/show how to read a file into memory...
 *
 */

/* don't try this sort of direct hardware access at home! */
static   xdata at 0xC000 char g_ucFileCacheBlockC;
volatile xdata at 0xFF18 unsigned int g_uiBlockC;
/* No Paul, I won't try this @ home! */


char AtapiFileCache(unsigned char p_ucDevice,
                    unsigned long p_ulExtent,
                    unsigned long p_ulLength)
{
  #define K4 4096L
  static xdata unsigned int               uiMemBlockCount;
  static xdata AtapiPacketBufferType      aucCdb;
  char                                    bResult = false;
  xdata unsigned char                     aucLBA[3];
  static xdata AtapiFileCacheEntryTypeSid sidEntry;
  static xdata AtapiFileCacheEntryTypeSid sidFileCacheCur = 0;
  static xdata simm_id                    sid4K;
  static xdata AtapiFileCacheEntryType   *pCur;
  static xdata void                      *p4K;
  static xdata AtapiFileCacheEntryTypePtr pEntry;
  static xdata unsigned int               uiCount;
  static xdata unsigned long              ulFileCacheTicks;
  static xdata unsigned long              ulFileCacheTimeCount;
  static xdata unsigned int               blk4K;
  volatile char                           bInit = true;
  
  printf("Cache+++\r\n");
  bInit = true;
  
  uiMemBlockCount = (unsigned int)(unsigned long)((p_ulLength + (K4) / (K4)));
  
//  if (uiMemBlockCount > 1536)
//    uiMemBlockCount = 1536;	// == 6MB file
  // Replaced by line below (thanx). This one works...  
  (uiMemBlockCount > 1536) ? (uiMemBlockCount = 1536) : (uiMemBlockCount = uiMemBlockCount);
  
  memset(aucCdb, 0, sizeof(aucCdb));
  aucCdb[0] = cmdATAPI_READ;	// read command code
  aucCdb[9] = 2;		// # of sectors to read (2 * 2K)

//  AtapiFileUncache(p_ucDevice);
  sidFileCacheCur    = 0; //m_sidFileCacheList[p_ucDevice];
  
  ulFileCacheTimeCount = clock_tick();

  for (uiCount = 0; uiCount < uiMemBlockCount; uiCount++)
  {
    // Use low-level memory allocation to be able to allocate a 4K block
    // in which to ATAPI sectors can be stored, each 2K in size...
    blk4K      = malloc_blocks(1);
    // Map block in the 0xC000 region
    g_uiBlockC = blk4K;
    
    // Setup the command data buffer to read the two sectors
    _Long2Char(p_ulExtent, &aucLBA);
    aucCdb[2] = aucLBA[3];
    aucCdb[3] = aucLBA[2];
    aucCdb[4] = aucLBA[1];
    aucCdb[5] = aucLBA[0];
    p_ulExtent += 2;

    bResult = AtapiSendCommand(p_ucDevice, sizeof(aucCdb), aucCdb, 4096, &g_ucFileCacheBlockC);
    
    sidEntry               = simm_malloc(sizeof(AtapiFileCacheEntryType));
    if (sidEntry == 0)
      printf("sid invalid\r\n");
    pEntry                 = (AtapiFileCacheEntryTypePtr)addr7(sidEntry);
    pEntry->sidNext        = 0;
    pEntry->blkMemoryBlock = blk4K;
    printf("CacheBlk = "); print_hex16(blk4K); print_crlf();
    
    //    if (0 == sidFileCacheCur)
    // Replaced by the line below (again thanx guys)
    if (bInit)
    {
      // Setting bInit to false directly does NOT work. Replaced by 2 stupid lines. Now the compiler
      // seems to remember to set bInit to false...
      bInit = true;
      bInit = false;
      sidFileCacheCur                = sidEntry;
      m_sidFileCacheList[p_ucDevice] = sidEntry;
      printf("1st cache = "); print_hex32(sidFileCacheCur); printf("blk = "); print_hex16(blk4K); print_crlf();
    }
    else
    {
      pCur            = addr6(sidFileCacheCur);
      pCur->sidNext   = sidEntry;
      sidFileCacheCur = sidEntry;
      printf("more cache = "); print_hex32(sidFileCacheCur); printf("blk = "); print_hex16(blk4K); print_crlf();
    }
  }

  ulFileCacheTicks = (clock_tick() - ulFileCacheTimeCount);
  printf("ATAPI-FileCache, len = "); print_hex32(p_ulLength);
  printf("4K = %u", uiMemBlockCount);
  printf("in %x"); print_hex32(ulFileCacheTicks); printf(" ticks ***\r\n");
}

#pragma RESTORE
