#define ATAPI_TEST
#pragma NOGCSE
/*
 * MP3 Player, Main Program, http://www.pjrc.com/tech/mp3
 * Copyright (c) 2000, PJRC.COM, LLC
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a specific exception to the GPL, the executable object code built
 * from this source code may be combined with hardware configuration data
 * files.  A hardware configuration data file is a set of data that is
 * transmitted to an intergrated circuit that is not a general purpose
 * microprocessor or microcontroller, in order to establish its normal
 * operation.  The process of combining the executable ojbect code built
 * from the GPL licensed source code with the hardware configuration data
 * shall be considered an aggregation of another work not based on the
 * Program.  While the GPL does not restrict use of the program, any
 * use restriction associated with the hardware configuration data (for
 * example, that it only be used with particular hardware) shall apply
 * to the combined file which includes a copy of the hardware configuration
 * data.
 *
 * Contact: paul@pjrc.com
 */



#include "paulmon2.h"
#include "as31glue.h"
#include "startup.h"
#include "printf.h"
#include "display.h"
#include "parse.h"
#include "malloc.h"
#include "dirlist.h"
#include "playlist.h"
#include "sta013.h"
#include "stricmp.h"
#include "rand.h"
#include "id3.h"
#include "params.h"
#include "treedir.h"
#include "ata.h"
#include "atapi.h"

#include <8051.h>

#include "main.h"




void init_current_list_and_file(void);
char is_state_normal(unsigned char state);
char play_blocks(char mp3_fd);
void process_resume();
void save_resume_info();
void dram_test(void);


typedef enum {
   S_READ_DIR,
   S_CACHING,
   S_PLAYING_AND_CACHING,
   S_PLAYING,
   S_FINISH_PLAYING,
   S_PAUSED
} state_t;



xdata simm_id /* playlist_t */ current_file;



// When debugging, waiting for the drive to spin up can be a pain. Use
// the parameter editor to set the spin value to Never


bit dummy_bit;

volatile xdata unsigned char playing;
idata unsigned char ibuf[4];
xdata unsigned long current_file_size=0xFFFFFFFF;
char resume;
play_modes_t play_mode;
unsigned int playlist_index,track_index;

// We reset this from the sta013 before we start each song.  It's used in display.c
// to display elapsed time. -ZSB 12-Aug-2001 3:15 PM
unsigned long songStartFrameCount=0;
// -ZSB 18-May-2002 3:50 PM: update_rand_needed changed to int to keep track of how many 
// times the seed has changed w/o being saved.  See get_move_count() for full comment on this.
unsigned char update_rand_needed; // Whether timer 3 updates the random seed
bit update_mode_needed;
bit update_sta013_needed;
bit do_id3; // Whether timer 3 scans the ID3 tag

void main()
{
	event_t event;
	char mp3_fd=-1;
	char r;
	volatile state_t state;
	volatile state_t prior_state;
	char caching_step;
	filelist_t * fl_struct;

	update_rand_needed = 0;
	do_id3 = 0;


	/* initialize everything */

	init_uart();	// from here forward, use print/printf

	print("\r\n\r\n\
\\[\\A0\\@!0PJRC MP3 Player 0.6.99\\]\r\n\
MP3 Player with ATAPI support, Version 0.6.99, June 16, 2002\r\n\
Copyright (C) 2002, PJRC.COM, LLC\r\n\r\n\
Type 'QUIT' to return to the monitor\r\n\r\n");

	// indicate EXPERIMENTAL RELEASE on relevant versions
	print("\\[\\B &EXPERIMENTAL DEV RELEASE\\]\r\n");

	printf("\\[\\C \\B 'Built %s\\]\r\n", __DATE__);

	playback.mode_known = 0;
	timer_setup();
	param_init();

	play_dma_irq_enable();

	print("\\[\\B !Waiting for Hard Drive\r\n\r\n\\]\r\n");
	while (!ide_init())
		;  // keep waiting until init is complete

	lcd_setup();

	parse_init();
	if (detect_filesystem() == 0) pm2_exit();

//	#MBA
	AtaQueryDevice();
    	AtapiStartStop(1, true);
    	AtapiReadTOC(1);
    	AtapiReadCapacity(1);
//    	#MBA

	restore_sta013_settings();
	update_sta013_needed = 0;

	debug = 0;

	print_memory_available();
        playlist_init();

	print("\\[\\B !Scanning directories    \\]\r\n");
	print("Building Playlist(s): Scan All Directories:\r\n");
	dirlist_build_init();
//	dirlist_build_ata(0);
	dirlist_build_atapi(1);
	dirlist_build_finish();
	treedir_init(); // must be done after root defined

//	AtapiEject(1);

	print_memory_available();

	// these print lots of info, but they take a very long
	// time to print if the drive has many gigs of files.
	//dirlist_show();
	dirlist_show_atapi();

	print("\\[\\B \"Building M3U playlists  \\]\r\n");
	m3u_list_show();
	//process_m3u();
	// playlist_show_all();


	init_current_list_and_file();

	restore_rand_settings();

#if 0
	current_playlist = first_playlist;
	playlist_show();
	randomize_one_playlist(current_playlist);
	playlist_show_rand();
	backup_rand_settings();
#endif


	/* This is the code for a simple resume function. Since
	   there is no way to set resume to 1 via the display, this
	   feature will be off unless explicitly turned on in the
	   code. This is just as well, because at the moment it breaks
	   if the playlist- or prev buttons are used until they cause 
	   a rollover. Im thinking about the best way to resolve this issue. */
	resume = param_value[PARAM_RESUME];
	//resume=1;
	playlist_index=0;
	track_index=0;

	if(resume) {
	    process_resume();
	}

	play_mode = param_value[PARAM_PLAY_MODE];
	if (play_mode == MODE_ALLRAND || play_mode == MODE_DIRRAND
	   || play_mode == MODE_PLSTRAND) {
		// -ZSB 12-Aug-2001 12:17 AM:
		// Since we always start at the first file of the first
		// playlist, we'll always start with the same song
		// regardless of whether we've retrieved a random mode
		//  from NV storage or not.  The next line forces us to
		// jump forward one track on startup if the saved mode
		// was random.			
		//
		// NOTE: If the conditions that determine a random state
		// ever change (IE in pl_next()), please be sure to update
		// those condtions here.
		//		
		// -ZSB 18-May-2002 3:28 PM: Modified to always do
		// pl_l_next, regardless of what type of random mode
		// we were in.

		// pl_l_next();
	} 	

	// random mode is broken right now (31-MAY-02)
	play_mode = MODE_SEQNTIAL; 


	update_mode_needed = 0;

	playing = 1;

	print("\r\nMP3 Player, Begin Playback\r\n");

	lcd_change_menu(0);
	init_silent_mp3_clip();
	lcd_display_id3_init();

	/* this infinite loop runs all the actions of the MP3 player */

	state = prior_state = S_READ_DIR;

	while (1) {
		/* first, process any events from the user interface */
		parse_serial_input();
		event = get_next_event();		
		switch (event) {
		  case E_NOTHING: {
			break;
		  }
		  case E_PLAY_PAUSE: {			  
			if (state == S_PAUSED) {
				print("Play\r\n");
				state = prior_state;
				play_resume();
				playing = 1;
				lcd_change(PARAM_PLAY_MODE);
				// lcd_play_pause_change();
			} else {
				if (is_state_normal(state)) {
					print("Pause\r\n");
					prior_state = state;
					state = S_PAUSED;
					play_suspend();
					playing = 0;
					lcd_change(PARAM_PLAY_MODE);
					// lcd_play_pause_change();
				} else {
					put_back_event(event);
				}
			}
			break;
		  }

		  case E_NEXT:
			if (is_state_normal(state)) {
				print("Next File\r\n");
				mute_sta013_output();
				play_abort();
				ide_flush();
				current_file = next_file(current_file);
				play_silent_mp3_clip();
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;
			
		  case E_PREV:
			if (is_state_normal(state)) {
				print("Previous File\r\n");
				mute_sta013_output();
				play_abort();
				ide_flush();
				current_file = prev_file(current_file);
				play_silent_mp3_clip();
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;

		   case E_VOL_UP:
                        action_vol_up();
                        break;
	 
                   case E_VOL_DOWN:
                        action_vol_down();
                        break;

		  case E_NEXT_LIST:
			if (is_state_normal(state)) {
				print("Next Play List\r\n");
				mute_sta013_output();
				play_abort();
				ide_flush();
				current_file = next_playlist(current_file);
				play_silent_mp3_clip();
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;
			
		  case E_PREV_LIST:
			if (is_state_normal(state)) {
				print("Previous Play List\r\n");
				mute_sta013_output();
				play_abort();
				ide_flush();
				current_file = prev_playlist(current_file);
				play_silent_mp3_clip();
				state = S_FINISH_PLAYING;
			} else {
				put_back_event(event);
			}
			break;

		  case E_RANDOM:
		        random_mode_next();
			lcd_change(PARAM_PLAY_MODE);		    
		        break;


		  case E_LCD_SCROLL_TIMER:
			set_timer(TIMER_LCD_SCROLL, SCROLL_SPEED);
			lcd_timer_callback();
			if (state == S_PLAYING_AND_CACHING || state == S_PLAYING) {				
				lcd_screen_dynamic(0);
			}
			break;		  
		  
		  case E_PARAM_WRITE_TIMER:
			// Any time the various NV parameters are changed,
			// set timer 3 to a reasonable timeout
			// (like 100 = 10 seconds).  That way, if the user
			// is cycling through an option a bunch of times,
			// hitting the volume buttons repeatedly, etc.,
			// we should reduce the number of writes to the
			// flash by only saving them every 10 seconds,
			// not for every single keypress.  If we lose power
			// within the 10 second timeout, the new settings
			// won't be saved, but that's hardly the end of the
			// world...   -ZSB 11-Aug-2001 11:39 PM
			
			eparam_write_flash();

			if (update_sta013_needed) {
				backup_sta013_settings();
				update_sta013_needed = 0;
			}

			if (update_mode_needed) {
		        	ibuf[0] = play_mode;
		        	write_flash_param(PARAM_PLAY_MODE, ibuf);
				update_mode_needed = 0;
			}

			if (update_rand_needed) {				
				// First 8
				backup_rand_settings();
				update_rand_needed = 0;
			}
			break;

		  case E_ID3_UPDATE_TIMER:
			if (do_id3) {
			  process_id3(mp3_fd, current_file_size);
			  do_id3 = 0;
			}
			break;

		  default:
			lcd_user_action(event);
			break;
		}

		do_ide_loop(6);

		/* and then do the next step to actually play MP3 files */


#ifndef NEW_STATE_MACHINE

		switch(state) {
		  case S_READ_DIR:
			lcd_set_playlistname(((playlist_list_t *)addr6(((playlist_t *)addr6(current_file))->playlist))->name);

			fl_struct = addr5(((playlist_t *)addr6(current_file))->fl_rec);
			current_file_size = fl_struct->size;
			lcd_set_filename(fl_struct->long_name);  // uses addr7
#ifdef ATAPI_TEST
                        // Fill with dummy value. Not used by simple ATAPI implementation
                        // Always one file open max
                        mp3_fd = 1;
#else
			mp3_fd = file_open_by_1st_cluster(fl_struct->cluster);
			file_cache(mp3_fd, 0, current_file_size);
#endif
			state = S_CACHING;
			caching_step = 0;
			r = SP;
			printfd("S_READ_DIR, mp3_fd=%d %x\r\n", mp3_fd, r);
			songStartFrameCount = GetFrameCount();
			play_blocks(-1);
			set_timer(TIMER_LCD_SCROLL, SCROLL_SPEED);
			break;

		  case S_CACHING:
			r = SP;
			printfd("S_CACHING, mp3_fd=%d %x\r\n", mp3_fd, r);
#ifdef ATAPI_TEST
			// Cache file. This might take some time, as the current implementation
			// uses PIO mode to get the file from CD. So be patient...
			printf("ATAPI-FileCache: Be patient, this might take some time...\r\n");
			AtapiFileCache(1, fl_struct->cluster, current_file_size);
			state = S_PLAYING_AND_CACHING;
#else
			r = file_cache_work(mp3_fd);
			/* TODO: check return value from file_cache_work */
			if (++caching_step > 6) {
				file_seek(mp3_fd, 0);
				state = S_PLAYING_AND_CACHING;
			}			
#endif
			break;

		  case S_PLAYING_AND_CACHING:
			printfd("S_PLAYING_CACHING, mp3_fd=%d\r\n", mp3_fd);
#ifdef ATAPI_TEST
			// Its simple. File is cached already.
			// Set new state, and do some id3 stuff (I do NOT know what...)
			AtapiFileReadInit(1);
			state  = S_PLAYING;
			do_id3 = 1;
			clear_timer(TIMER_ID3_UPDATE);
			set_timer(TIMER_ID3_UPDATE, 50);				
#else
			r = file_cache_work(mp3_fd);
			if (r) {
				state = S_PLAYING;
				if (param_value[PARAM_DISK_SPIN_DELAY] != 0) {
				   ide_sleep();
				}
				do_id3 = 1;
				clear_timer(TIMER_ID3_UPDATE);
				set_timer(TIMER_ID3_UPDATE, 50);				
				// process_id3(mp3_fd, current_file_size);
				free_fat_memory();
			}
#endif
			/* fall through into S_PLAYING's code */

		  case S_PLAYING:
			printfd("S_PLAYING\r\n");
			if (play_blocks(mp3_fd) != 0) {
				current_file = next_file(current_file);
				play_silent_mp3_clip();
				state = S_FINISH_PLAYING;
			}
			break;

		  case S_FINISH_PLAYING:
			// Resettings the songStartFrameCount here should
			// get us a new count at the start of each song
			printf("S_FINISH_PLAYING, q=\r\n", play_queue_avail());
			if (is_play_queue_empty() && is_ide_idle())
			{
                                lcd_display_id3_init();
#ifdef ATAPI_TEST
				AtapiFileUncache(1);
				current_file_size = 0xFFFFFFFF;
#else
				file_close(mp3_fd);
				current_file_size = 0xFFFFFFFF;
				free_fat_memory();
#endif
				print_memory_available();
				if (is_ide_sleeping())
				{
					ide_hard_reset();
					while (!ide_init())
						; // wait until init complete
				}
				state = S_FINISH_PLAYING;
				printf("State set to S_READ_DIR\r\n");
				state = S_READ_DIR;
			}
			songStartFrameCount = GetFrameCount();
			break;

		  case S_PAUSED:
			if (prior_state == S_PLAYING_AND_CACHING) {
				r = file_cache_work(mp3_fd);
				if (r) {
					prior_state = S_PLAYING;
					if (param_value[PARAM_DISK_SPIN_DELAY] != 0) {
					   ide_sleep();
					}
					do_id3 = 1;
					clear_timer(TIMER_ID3_UPDATE);
					set_timer(TIMER_ID3_UPDATE, 50);
					// process_id3(mp3_fd, current_file_size);
					free_fat_memory();
				}
			 }
			break;

		  default:			
		    print("ILLEGAL STATE:");
			print_hex16(state);
			print_crlf();
		}


#else //NEW_STATE_MACHINE


		// This state machine will control playing data from memory

		switch (play_state) {
		  case S_START_NEW_FILE:


		  case S_PLAYING_FILE:


		  case S_FINISH_PLAYING_FILE:


		  case S_FINISHED_FILE:


		  case S_PAUSED:
			break;

		  default:
			print("Illegal play_state\r\n");
			play_state = S_START_NEW_FILE;
			break;
		}


		// And this separate state machine will be responsible for
		// disk access.  The idea is to be able to do various things
		// with the disk drive independent of what the current state
		// of the playback might happen to be.

		switch (disk_state) {
		  case S_DRIVE_INACTIVE:
			if (disk_activity_needed) { // TODO && sleep not pending
				ide_hard_reset();
				disk_time_mark = clock_tick();
				disk_state = S_WAIT_DRIVE_READY;
			}
			break;
		  case S_WAIT_DRIVE_READY:
			if (ide_init()) {
				disk_latency = (clock_tick() - disk_time_mark);
				disk_state = S_CACHING_PRIORITY_1;
				caching_priority = 1;
			} else {
				// TODO, timeout, go back to and try again
			}
			break;

		  case S_CACHING:
			if (highest_priority == NULL) {
				ide_sleep();
				free_fat_memory();
				disk_state = S_WAIT_SHUTDOWN;
				break;
			}
			r = file_cache_work(highest_priority->file_desc);
			if (r == 1) {  // finished caching this file
				highest_priority->priority = 0;
				highest_priority = highest_priority->next;
				break;
			}
			if (r == 2) {  // ran out of memory
				highest_priority = NULL;
			}
			break;

		  case S_WAIT_SHUTDOWN:
			// TODO, check if drive is actually in sleep mode
			// and remain in this state until it really is

			disk_state = S_DRIVE_INACTIVE;
			break;

		  default:
			print("Illegal disk_state\r\n");
			play_state = S_DRIVE_INACTIVE;
			break;
		}

#endif //NEW_STATE_MACHINE

	}

	pm2_exit();
}





/* choose "current_file" for initial startup */

void init_current_list_and_file(void)
{
	simm_id /* playlist_list_t */ pl;
	playlist_list_t *pl_ptr;

	pl=first_playlist;
	while (pl != 0) {
		pl_ptr = addr6(pl);
		if (pl_ptr->playlist_start_sequential != 0) {
			current_file = pl_ptr->playlist_start_sequential;
			return;
		}
		pl = pl_ptr->next_sequential;
	}

	print("\\[\\B \"   - No files found -   \\]\r\n");
	print("No files found...\r\n");
	pm2_exit();
}





char is_state_normal(unsigned char state)
{
	if (state == S_PLAYING
	 || state == S_PLAYING_AND_CACHING
	 || state == S_FINISH_PLAYING) {
		return 1;
	}
	return 0;
}

#define BLOCK_SIZE 4096


/* return 0 if there is more to play, or non-zero if we have played */
/* everything or nothing more can play for some reason */

char play_blocks(char mp3_fd)
{
	unsigned int block;
	static unsigned long played_bytes;
	static unsigned char eof_count;

	if (mp3_fd == -1) {
		played_bytes = 0;
		eof_count = 0;
		return 1;
	}

	if (current_file_size == 0xFFFFFFFF) {
		print("WARNING: play_blocks called without known file size\r\n");
		return 1;
	}

	while (play_queue_avail()) {
#ifdef ATAPI_TEST
		block = AtapiFileReadBlock(1);
#else
		block = file_read_block(mp3_fd);
#endif
		printf("FRB, r=%x, played=%lx, size=%lx\r\n",
			block, played_bytes, current_file_size);
		if (block == 0) {
			return 0;	// not in cache yet
		}
		if (block == 0xFFFF) {
			if (++eof_count > 240) {
				played_bytes = 0;	// end of file (physical on disk)
				return 1;
			} else {
				return 0;		// maybe more will be read soon?
			}
		}
		if (current_file_size - played_bytes > BLOCK_SIZE) {
			play_block(block, BLOCK_SIZE);
			played_bytes += BLOCK_SIZE;
		} else {
			play_block(block, current_file_size - played_bytes);
			played_bytes = current_file_size;
			return 1;
		}
	}
	eof_count = 0;
	return 0;
}





void process_resume()
{
  unsigned int count;
  //unsigned int i;

  count = ((int)param_value[PARAM_PLST_INDEX_HI] << 8) 
     + param_value[PARAM_PLST_INDEX_LO];
  
  //printf("playlist: %d \r\n",count);

  //Use sequential mode so that the indices count forward right.
  play_mode = MODE_SEQNTIAL; 
  resume=0;

#if 0
  if(count > 0)
    {
      for(i=1;i<=count;i++)
	{
	  pl_l_next();
	}
    }

  count = ((int)param_value[PARAM_TRACK_INDEX_HI] << 8) +
     param_value[PARAM_TRACK_INDEX_LO];
  //printf("track: %d \r\n",count);
  if(count > 0)
    {
      for(i=1;i<=count;i++)
	{
	  pl_next();
	}
    }
#endif

  resume=1;
}

void save_resume_info()
{

  unsigned char hi,lo;
  //printf("playlist_index: %d track_index: %d\r\n",playlist_index,track_index);
  hi=(playlist_index & 0xFF00)>>8;
  lo=(playlist_index & 0x00FF);
  
  // FIXME: I'm leaving these flash writes outside of the timer
  // until the resume functionality is a little more settled.
  // Ideally, this should go to disk or something anyways...
  // -ZSB 11-Aug-2001 11:30 PM
  ibuf[0]=lo;
  write_flash_param(PARAM_PLST_INDEX_LO, ibuf);
  ibuf[0]=hi;
  write_flash_param(PARAM_PLST_INDEX_HI, ibuf);

  hi=(track_index & 0xFF00)>>8;
  lo=(track_index & 0x00FF);
  ibuf[0]=lo;
  write_flash_param(PARAM_TRACK_INDEX_LO, ibuf);
  ibuf[0]=hi;
  write_flash_param(PARAM_TRACK_INDEX_HI, ibuf);

}
